////////////////////////////////////////////////////////////////////////
// Class:       FrequencyStatisticalTestRoiFinder
// Plugin Type: producer (art v3_05_01)
// File:        FrequencyStatisticalTestRoiFinder_module.cc
//
// Generated at Wed Nov 18 15:26:59 2020 by Wanwei Wu using cetskelgen
// from cetlib version v3_10_00.
// About: waveform ROI finder based on statistical test using frequency
//   a. option to use one of two methods: chi-square (chi2) test and Delta chi-square (dchi2) test
//      chi2 test: using only noise frequency template
//      dchi2 test: using both noise frequency template and signal frequency template
//   b. frequency bins for chi2 or dchi2: first fNumFrequencyBins 
//   c. options for window size and stride size; a shift to last window in order to have the same window size
//   d. after filling true for roi, filling false for non-roi will narrow
//      the roi range.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib_except/exception.h"
#include "cetlib/search_path.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art_root_io/TFileService.h"
#include "art_root_io/TFileDirectory.h"

// larsoft includes
#include "lardataobj/RawData/raw.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardataobj/RecoBase/Wire.h"
#include "lardata/ArtDataHelper/WireCreator.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "larcore/Geometry/Geometry.h"

#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"

#include <memory>

#include "TFile.h"
#include "TH1D.h"
#include "TProfile.h"

using namespace std;


class FrequencyStatisticalTestRoiFinder;


class FrequencyStatisticalTestRoiFinder : public art::EDProducer {
public:
  explicit FrequencyStatisticalTestRoiFinder(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  FrequencyStatisticalTestRoiFinder(FrequencyStatisticalTestRoiFinder const&) = delete;
  FrequencyStatisticalTestRoiFinder(FrequencyStatisticalTestRoiFinder&&) = delete;
  FrequencyStatisticalTestRoiFinder& operator=(FrequencyStatisticalTestRoiFinder const&) = delete;
  FrequencyStatisticalTestRoiFinder& operator=(FrequencyStatisticalTestRoiFinder&&) = delete;

  // Required functions.
  void produce(art::Event& e) override;

private:
  // user's function
  void beginJob() override;
  std::vector<bool> findROI(const std::vector<float>& adcin, int view);

  // configuration from fhicl file
  std::string fRawDigitLabel;
  std::string fWireProducerLabel;
  bool fUseShort; // false: float for ADC; true: short for ADC
  bool fUseDChi2; // true: use dchi test; false: use chi2 test
  std::vector<float> fChi2Cut;
  std::vector<float> fDChi2Cut;
  int fWindowSize;
  int fStrideSize;
  std::string fFrequencyTemplateFile;
  int fNumFrequencyBins;
  int fxNumBins;
  float fxBinWidth;

  int fNticks;
  int fNumStrides;
  int fLastWindowSize;
  int fNPlanes;
  float fFFTBinWidth;

  TProfile* pf_frequency_noise_template[3];
  TProfile* pf_frequency_signal_template[3];
};


FrequencyStatisticalTestRoiFinder::FrequencyStatisticalTestRoiFinder(fhicl::ParameterSet const& p)
  : EDProducer{p}  // ,
  // More initializers here.
{
  cet::search_path sp("FW_SEARCH_PATH");

  fRawDigitLabel           = p.get<std::string>("RawDigitLabel");
  fWireProducerLabel       = p.get<std::string>("WireProducerLabel");
  fUseShort                = p.get<bool>("UseShort", true);
  fUseDChi2                = p.get<bool>("UseDChi2", true);
  fChi2Cut                 = p.get<std::vector<float>>("Chi2Cut");
  fDChi2Cut                = p.get<std::vector<float>>("DChi2Cut");
  fWindowSize              = p.get<int>("WindowSize", 200);
  fStrideSize              = p.get<int>("StrideSize", 150);
  fFrequencyTemplateFile   = sp.find_file(p.get<std::string>("FrequencyTemplateFile"));
  fNumFrequencyBins        = p.get<int>("NumFrequencyBins", 50);
  fxNumBins                = p.get<int>("xNumBins", 100);
  fxBinWidth               = p.get<float>("xBinWidth", 0.02);

  if (fRawDigitLabel.empty() && fWireProducerLabel.empty()){
    throw cet::exception("FrequencyStatisticalTestRoiFinder") << "Both RawDigitLabel and WireProducerLabel are empty";
  }

  if ((!fRawDigitLabel.empty()) && (!fWireProducerLabel.empty())){
    throw cet::exception("FrequencyStatisticalTestRoiFinder") << "Only one of RawDigitLabel and WireProducerLabel should be set";
  }

  // DetectorPropertiesService
  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataForJob();
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob();
  fNticks = detProp.ReadOutWindowSize(); // number of clock ticks per event
  float fSampleRate = sampling_rate(clockData); // period of the TPC readout electronics clock [us]; 1 tick = 1 tdc = 198 ns.
  fFFTBinWidth = 1.0/(fNticks*fSampleRate*1.0e-6); // [kHz]
  //std::cout << "Number of clock ticks per event: " << fNticks << std::endl;

  produces< std::vector<recob::Wire> >();

  if (fWindowSize > 0 && fStrideSize >0) {
    float dmn = fNticks - fWindowSize; // distance between tail edge of 1st window and last data point
    fNumStrides = std::ceil(dmn / float(fStrideSize));
    unsigned int overshoot  = fNumStrides * fStrideSize + fWindowSize - fNticks;
    fLastWindowSize = fWindowSize - overshoot;
    unsigned int numwindows = fNumStrides + 1;
    std::cout << " !!!!! FrequencyStatisticalTestRoiFinder: WindowSize = " << fWindowSize
              << ", StrideSize = " << fStrideSize
              << ", dmn/StrideSize = " << dmn / fStrideSize << std::endl;
    std::cout << "   dmn = " << dmn << ", NumStrides = " << fNumStrides
              << ", overshoot = " << overshoot << ", LastWindowSize = " << fLastWindowSize
              << ", numwindows = " << numwindows << std::endl;
  }

  fNPlanes = lar::providerFrom<geo::Geometry>()->Nplanes();
  if (fNPlanes > 3) {
    throw cet::exception("FrequencyStatisticalTestRoiFinder") << "Number of planes is " << fNPlanes << "; please check";
  }
}


void FrequencyStatisticalTestRoiFinder::produce(art::Event& e)
{
  auto const* geo = lar::providerFrom<geo::Geometry>();

  // get RawDigit
  art::Handle< std::vector<raw::RawDigit> > rawdigitListHandle;
  std::vector< art::Ptr<raw::RawDigit> > rawdigitlist;
  if (e.getByLabel(fRawDigitLabel, "", rawdigitListHandle)) {
    art::fill_ptr_vector(rawdigitlist, rawdigitListHandle);
  }
  
  // get Wire
  art::Handle< std::vector<recob::Wire> > wireListHandle;
  std::vector<art::Ptr<recob::Wire> > wirelist;
  if (e.getByLabel(fWireProducerLabel, wireListHandle)) {
    art::fill_ptr_vector(wirelist, wireListHandle);
  }

  // make a collection of Wires and an association set
  std::unique_ptr<std::vector<recob::Wire> > wirecol(new std::vector<recob::Wire>);

  // loop over wires
  // use one of rawdigitlist and  wirelist that is not empty()
  for (unsigned int ich=0; ich < (rawdigitlist.empty()?wirelist.size():rawdigitlist.size()); ++ich) {

    std::vector<float> adcvec(fNticks);
    int view = -1;

    if (!wirelist.empty()) {
      const auto & wire = wirelist[ich];
      const auto & signal = wire->Signal();
      
      view = wirelist[ich]->View();

      if (fUseShort) {
        for (size_t j=0; j<adcvec.size(); ++j) {
          adcvec[j] = short(signal[j]);
        }
      }
      else {
        for (size_t j=0; j<adcvec.size(); ++j) {
          adcvec[j] = signal[j];
        }
      }
    }
    else if (!rawdigitlist.empty()) {
      const auto & digitVec = rawdigitlist[ich];
  
      view =  geo->View(digitVec->Channel());

      std::vector<short> rawadc(fNticks);
      raw::Uncompress(digitVec->ADCs(), rawadc, digitVec->GetPedestal(), digitVec->Compression());

      for (size_t j=0; j<rawadc.size(); ++j) {
        adcvec[j] = rawadc[j] - digitVec->GetPedestal(); 
      }
    }

    // Statistical test using freqeuncy for Roi Finder of each window
    std::vector<bool> inroi(fNticks, false);
    
    inroi = findROI(adcvec, view);

    recob::Wire::RegionsOfInterest_t rois(fNticks);

    std::vector<float> sigs;
    int lastsignaltick = -1;
    int roistart = -1;

    for (int j = 0; j < fNticks; ++j) {
      if (inroi[j]) {
        if (sigs.empty()) {
          sigs.push_back(adcvec[j]);
          lastsignaltick = j;
          roistart = j;
        }
        else {
          if (j != lastsignaltick + 1) {
            rois.add_range(roistart, std::move(sigs));
            sigs.clear();
            sigs.push_back(adcvec[j]);
            lastsignaltick = j;
            roistart = j;
          }
          else {
            sigs.push_back(adcvec[j]);
            lastsignaltick = j;
          }
        }
      }
    }   
    if (!sigs.empty()) { rois.add_range(roistart, std::move(sigs)); }

    if (!wirelist.empty()) {
      wirecol->emplace_back(recob::Wire(rois, wirelist[ich]->Channel(), wirelist[ich]->View()));
    }
    else if (!rawdigitlist.empty()) {
      wirecol->emplace_back(recob::Wire(rois, rawdigitlist[ich]->Channel(), geo->View(rawdigitlist[ich]->Channel())));
    }

  } // loop over ich

  e.put(std::move(wirecol));
}


void FrequencyStatisticalTestRoiFinder::beginJob() {
  // frequency template
  TFile TemplateFFTFile(fFrequencyTemplateFile.c_str());
  for (int i=0; i<fNPlanes; i++) {
    if (fUseShort) {
      pf_frequency_noise_template[i] = (TProfile*)TemplateFFTFile.Get(Form("pf_frequency_noise_short_%d",i));
      pf_frequency_signal_template[i] = (TProfile*)TemplateFFTFile.Get(Form("pf_frequency_signal_short_%d",i));
    }
    else {
      pf_frequency_noise_template[i] = (TProfile*)TemplateFFTFile.Get(Form("pf_frequency_noise_float_%d",i));
      pf_frequency_signal_template[i] = (TProfile*)TemplateFFTFile.Get(Form("pf_frequency_signal_float_%d",i));
    }
  }
}


std::vector<bool> FrequencyStatisticalTestRoiFinder::findROI(const std::vector<float>& adcin, int view) {
  std::vector<bool> roivec(fNticks, false);
  if (int(adcin.size()) != fNticks) {return roivec;}

  // create a vector of windows
  std::vector<std::vector<float>> windowvec(fNumStrides + 1, std::vector<float>(fWindowSize, 0.));

  // fill each window with adc values
  unsigned int j1, j2, k0;
  for (int i = 0; i < fNumStrides; i++) {
    j1 = i * fStrideSize;
    j2 = j1 + fWindowSize;
    k0 = 0;
    for (unsigned int j = j1; j < j2; j++) {
      windowvec[i][k0] = adcin[j];
      k0++;
    }
  }
  // last window as a special case: still has a size of fWindowSize, by shifting backward.
  // Another possible is to fill remaining elements with 0 to maintain the fWindowSize, which is not considered here.
  j2 = fNticks;
  j1 = j2 - fWindowSize;
  k0 = 0;
  for (unsigned int j = j1; j < j2; j++) {
    windowvec[fNumStrides][k0] = adcin[j];
    k0++;
  }

  // set to true for all bins in a window if it is identified as signal
  for (int i = 0; i < fNumStrides + 1; i++) {
    
    TH1D *h1_waveform = new TH1D("h1_waveform", "h1_waveform", fWindowSize, 0, fWindowSize);
    for (int k = 0; k < fWindowSize; ++k) {
      h1_waveform->SetBinContent(k+1, windowvec[i][k]);
    }
    
    TH1D *h1_fft = new TH1D("h1_fft", "h1_fft", fWindowSize, 0, fWindowSize*fFFTBinWidth);
    
    TH1* h1_transform = 0;
    h1_transform = h1_waveform->FFT(h1_transform, "MAG");

    for (int k = 0; k < fWindowSize; ++k) {
      double frequency = (k+0.5) * fFFTBinWidth; // [kHz]
      double amplitude = h1_transform->GetBinContent(k+1);
      h1_fft->Fill(frequency, amplitude);
    }

    if (fUseDChi2) {
      // dchi2 according to template: dchi2 = chi2_n - chi2_{n+x*s}. Here, n for noise; s for signal; best x gives the minimum of chi2_{n+x*s}, hence dchi2 will be the maximum. Also, a scale factor of 1./fNumFrequencyBins is introduced for dchi2 for convenience.
      double maxdchi2 = -1e20;
      //double x = 0.;
      for (int m = 0; m < fxNumBins; m++) {
        double tempdchi2 = 0.;
        for (int k = 0; k < fNumFrequencyBins; ++k) {
          double value_noise = (h1_fft->GetBinContent(k+1) - pf_frequency_noise_template[view]->GetBinContent(k+1)) / (pf_frequency_noise_template[view]->GetBinError(k+1));
          double value_signal = (pf_frequency_signal_template[view]->GetBinContent(k+1)) / (pf_frequency_noise_template[view]->GetBinError(k+1));

          tempdchi2 += 2 * (m * fxBinWidth) * value_noise * value_signal - (m * fxBinWidth) * (m * fxBinWidth) * value_signal * value_signal;
        }

        if (tempdchi2 > maxdchi2) {
          maxdchi2 = tempdchi2;
          //x = m * fxBinWidth;
        }
      }   

      maxdchi2 = maxdchi2 / fNumFrequencyBins;
    
      if (i < fNumStrides) {
        if (maxdchi2 > fDChi2Cut[view]) {
          std::fill_n(roivec.begin() + i * fStrideSize, fWindowSize, true);
        }
        // filling false here can narrow the roi range
        else {
          std::fill_n(roivec.begin() + i * fStrideSize, fWindowSize, false);
        }
      }
      // last window
      else {
        if (maxdchi2 > fDChi2Cut[view]) {
          std::fill_n(roivec.begin() + fNticks - fWindowSize, fWindowSize, true);
        }
        // filling false here can narrow the roi range
        else {
          std::fill_n(roivec.begin() + fNticks - fWindowSize, fWindowSize, false);
        }
      }
    } // if (fUseDChi2)
    else {
      double chi2 = 0.;
      for (int k = 0; k < fNumFrequencyBins; ++k) {
        double value_noise = (h1_fft->GetBinContent(k+1) - pf_frequency_noise_template[view]->GetBinContent(k+1)) / (pf_frequency_noise_template[view]->GetBinError(k+1));
        chi2 += value_noise*value_noise;
      }

      double chi2ndf = chi2 / fNumFrequencyBins;

      if (i < fNumStrides) {
        if (chi2ndf > fChi2Cut[view]) {
          std::fill_n(roivec.begin() + i * fStrideSize, fWindowSize, true);
        }
        // filling false here can narrow the roi range
        else {
          std::fill_n(roivec.begin() + i * fStrideSize, fWindowSize, false);
        }
      }
      // last window
      else {
        if (chi2ndf > fChi2Cut[view]) {
          std::fill_n(roivec.begin() + fNticks - fWindowSize, fWindowSize, true);
        }
        // filling false here can narrow the roi range
        else {
          std::fill_n(roivec.begin() + fNticks - fWindowSize, fWindowSize, false);
        }
      }
    } // use chi2

    delete h1_transform;
    delete h1_waveform;
    delete h1_fft;

  } // loop over stride i

  return roivec;

} // findROI

DEFINE_ART_MODULE(FrequencyStatisticalTestRoiFinder)
