////////////////////////////////////////////////////////////////////////
/// \file   SignalShapingServiceT962_service.cc
/// \author H. Greenlee, B. Baller ported to ArgoNeuT from LArIAT
////////////////////////////////////////////////////////////////////////

#include "SignalShapingServiceT962.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib_except/exception.h"
#include "larcore/Geometry/Geometry.h"
#include "larcorealg/Geometry/TPCGeo.h"
#include "larcorealg/Geometry/PlaneGeo.h"
#include "lardata/DetectorInfoServices/DetectorClocksService.h"
#include "lardata/Utilities/LArFFT.h"
#include "TFile.h"
#include <fstream>

//----------------------------------------------------------------------
// Constructor.
util::SignalShapingServiceT962::SignalShapingServiceT962(const fhicl::ParameterSet& pset,
                                                           art::ActivityRegistry& /* reg */)
: fInit(false)
, fColFilterFunc(nullptr)
, fIndFilterFunc(nullptr)
{
  reconfigure(pset);
}


//----------------------------------------------------------------------
// Destructor.
util::SignalShapingServiceT962::~SignalShapingServiceT962()
{
  if(fColFilterFunc) delete fColFilterFunc;
  if(fIndFilterFunc) delete fIndFilterFunc;
}


//----------------------------------------------------------------------
// Reconfigure method.
void util::SignalShapingServiceT962::reconfigure(const fhicl::ParameterSet& pset)
{
  // Reset initialization flag.
  
  fInit = false;
  
  // Reset kernels.
  
  for(auto& ssg : fSignalShapeGroup) ssg.Reset();
  
  cet::search_path sp("FW_SEARCH_PATH");
  sp.find_file(pset.get<std::string>("ElectronicsResponseFile"), fResponseFile);
  fLongTailGroupTolerance = pset.get<double>("LongTailGroupTolerance", -1);
  
  // Induction plane field parameterization and filter
  fIndFieldParams  = pset.get<std::vector<float>>("IndFieldParams");
  if(fIndFieldParams.size() < 3 || fIndFieldParams[0] <= 0) {
    std::cout<<"IndFieldParams configuration error\n";
    exit(1);
  }
  std::string indFilt = pset.get<std::string>("IndFilter");
  std::vector<double> indFiltParams = pset.get<std::vector<double>>("IndFilterParams");
  fIndFilterFunc = new TF1("indFilter", indFilt.c_str());
  for(size_t i=0; i<indFiltParams.size(); ++i) fIndFilterFunc->SetParameter(i, indFiltParams[i]);
  
  // Collection plane field parameterization and filter
  fColFieldParams  = pset.get<std::vector<float>>("ColFieldParams");
  if(fColFieldParams.size() != 2 || fColFieldParams[0] <= 0) {
    std::cout<<"ColFieldParams configuration error\n";
    exit(1);
  }
  std::string colFilt = pset.get<std::string>("ColFilter");
  std::vector<double> colFiltParams = pset.get<std::vector<double> >("ColFilterParams");
  fColFilterFunc = new TF1("colFilter", colFilt.c_str());
  for(size_t i=0; i<colFiltParams.size(); ++i) fColFilterFunc->SetParameter(i, colFiltParams[i]);

} // reconfigure


//----------------------------------------------------------------------
// Accessor for single-plane signal shaper.
const util::SignalShaping& util::SignalShapingServiceT962::SignalShaping(unsigned int channel) const
{
  if(!fInit) init();
  // find the index of the appropriate convolution kernel
  unsigned short grp = fChanGroup[channel];
  return fSignalShapeGroup[grp];
} // SignalShaping


//----------------------------------------------------------------------
// Initialization method.
// Here we do initialization that can't be done in the constructor.
// All public methods should ensure that this method is called as necessary.
void util::SignalShapingServiceT962::init()
{
  if(fInit) return;
  fInit = true;
  
  art::ServiceHandle<util::LArFFT> fft;
  fftSize = fft->FFTSize();
  // This function also defines the exponential tail decay constants - fTailConsts
  SetElectResponse();
  // Calculate field and electronics response functions.
  SetFieldResponse();
  SetFilters();
  fSignalShapeGroup.resize(fTailConsts.size());
  
  std::vector<bool> isDone(fTailConsts.size(), false);
  for(unsigned short chan = 0; chan < nChans; ++chan) {
    unsigned short grp = fChanGroup[chan];
    // already done?
    if(isDone[grp]) continue;
    auto& ssg = fSignalShapeGroup[grp];
    ssg.AddResponseFunction(fElectResponse);
    if(grp == 0) {
      // induction plane
      ssg.AddResponseFunction(fIndFieldResponse);
      ssg.AddFilterFunction(fIndFilter);
    } else {
      // collection plane
      ssg.AddResponseFunction(fColFieldResponse);
      ssg.AddFilterFunction(fColFilter);
      /*
       // TODO: Add long-tail response here
       if(fTailConsts[grp] < -1E-4) {
       
       }
       */
    }
    ssg.save_response();
    ssg.set_normflag(false);
    ssg.SetPeakResponseTime(0.);
    ssg.CalculateDeconvKernel();
    isDone[grp] = true;
  } // chan
  
} // init()


//----------------------------------------------------------------------
// Calculate field response.
void util::SignalShapingServiceT962::SetFieldResponse()
{

  fColFieldResponse.resize(fftSize, 0.);
  fIndFieldResponse.resize(fftSize, 0.);
  
  mf::LogInfo("SignalShapingServiceT962") << " using the old (Best!) field shape " ;
  unsigned short nbinc = std::nearbyint(fColFieldParams[0]);
  if(nbinc > fftSize) {
    std::cout<<"SetFieldResponse nbinc screw-up "<<nbinc<<" fftSize "<<fftSize<<"\n";
    exit(1);
  }
  
  double integral = 0.;
  for(unsigned short ii = 1; ii < nbinc; ++ii) {
    fColFieldResponse[ii] = ii * ii;
    integral += fColFieldResponse[ii];
  } // ii
  
  for(unsigned short ii = 0; ii < nbinc; ++ii) fColFieldResponse[ii] *= fColFieldParams[1] / integral;
  
  // now the induction plane
  unsigned short nbini = std::nearbyint(fIndFieldParams[0]);
  if(nbini > fftSize) {
    std::cout<<"SetFieldResponse nbini screw-up "<<nbini<<" fftSize "<<fftSize<<"\n";
    exit(1);
  }
  integral = 0;
  unsigned short lastbini = 2 * nbini;
  for(unsigned short ii = 0; ii < lastbini; ++ii) {
    double ang = ii * M_PI / nbini;
    fIndFieldResponse[ii] = sin(ang);
    if(fIndFieldResponse[ii] > 0) {
      integral += fIndFieldResponse[ii];
    } else {
      // scale the negative lobe
      fIndFieldResponse[ii] *= fIndFieldParams[2];
    }
  } // ii
  ++lastbini;
  for(unsigned short ii = 0; ii < lastbini; ++ii) fIndFieldResponse[ii] *= fIndFieldParams[1] / integral;
} // SetFieldResponse

//----------------------------------------------------------------------
void util::SignalShapingServiceT962::SetElectResponse()
{
  // Extract the electronics response from a root histogram. Extract the long-tail
  // exponential time constants from the same file and put similar values into groups. 
  // The group index for each channel is stored in fChanGroup. A convolution kernel will
  // later be created for each group.
  TFile f(fResponseFile.c_str());
  if( f.IsZombie() ) {
    std::cout<< "Cannot open response file "<< fResponseFile.c_str()<<"\n";
    exit(1);
  } // zombie
  TH1D* electResp = (TH1D*)f.Get("real/electronics");
  unsigned short nbins = electResp->GetNbinsX();
  fElectResponse.resize(fftSize, 0);
  double peak = 0;
  for(unsigned short bin = 0; bin < nbins; ++bin) {
    fElectResponse[bin] = electResp->GetBinContent(bin + 1);
    if(fElectResponse[bin] > peak) peak = fElectResponse[bin];
  }
  
  if(fLongTailGroupTolerance > 0) {
    // The first group is for the induction plane that doesn't have long tails
    fTailConsts.resize(1, 0);
    for(unsigned short chan = 0; chan < 240; ++chan) fChanGroup[chan] = 0;
    // extract the long exponential tail time constants for each channel in the collection plane
    TH1D* tailConst = (TH1D*)f.Get("real/decayHist");
    if(tailConst->GetNbinsX() != nChans) {
      std::cout<<"Number of channels in real/decayHist is not nChan "<<tailConst->GetNbinsX()<<"\n";
      exit(1);
    }
    for(unsigned short chan = 240; chan < nChans; ++chan) {
      double tc = tailConst->GetBinContent(chan + 1);
      // ignore group 0 (induction plane)
      unsigned short grp = 1;
      for(grp = 1; grp < fTailConsts.size(); ++grp) {
        if(std::abs(tc - fTailConsts[grp]) < fLongTailGroupTolerance) break;
      } // indx
      if(grp == fTailConsts.size()) fTailConsts.push_back(tc);
      fChanGroup[chan] = grp;
    } // chan
    mf::LogInfo logInfo("SignalShapingServiceT962");
    logInfo<<"Created "<<fTailConsts.size()<<" decay tail group time constants:";
    for(unsigned short grp = 0; grp < fTailConsts.size(); ++grp) logInfo<<" "<<fTailConsts[grp];
  } else {
    // no long exponential tails. Define two groups of kernels
    fTailConsts.resize(2, 0);
    for(unsigned short chan = 0; chan < 240; ++chan) fChanGroup[chan] = 0;
    for(unsigned short chan = 240; chan < nChans; ++chan) fChanGroup[chan] = 1;
  } // no long exponential tail

} // SetElectResponse


//----------------------------------------------------------------------
// Calculate filter functions.
void util::SignalShapingServiceT962::SetFilters()
{
  int n = 1 + fftSize / 2;

  // N.B. Clock data is either global (which can be updated on input
  //      file boundaries) or per-event.  Unless the interface is
  //      changed to accept a per-event clock-data object, we must use
  //      the global data.  In general, lazy initialization is fraught
  //      with thread-safety issues and should be avoided.
  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService const>()->DataForJob();
  
  // Calculate filters.
  fIndFilter.resize(n);
  fIndFilterFunc->SetRange(0, double(n));
  fColFilter.resize(n);
  fColFilterFunc->SetRange(0, double(n));
  double srn = sampling_rate(clockData) * n;
  for(int ii = 0; ii < n; ++ii) {
    double freq = 500. * ii / srn;      // Cycles / microsecond.
    double f = fIndFilterFunc->Eval(freq);
    fIndFilter[ii] = TComplex(f, 0.);
    f = fColFilterFunc->Eval(freq);
    fColFilter[ii] = TComplex(f, 0.);
  } // i
} // SetFilters


namespace util {

  DEFINE_ART_SERVICE(SignalShapingServiceT962)

}
